/*
 * AFsplitter - Anti forensic information splitter
 * Copyright 2004, Clemens Fruhwirth <clemens@endorphin.org>
 *
 * AFsplitter diffuses information over a large stripe of data,
 * therefor supporting secure data destruction.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <errno.h>

#include "charcrandom.h"
#include "blake2.h"
#include "af.h"

static inline void XORblock(const char *src1, const char *src2, char *dst, size_t n)
{
	size_t j;

	for (j = 0; j < n; j++)
		dst[j] = src1[j] ^ src2[j];
}

/* diffuse: Information spreading over the whole dataset with
 * the help of blake2b.  Return 0 on success.
 */
static int diffuse(char *src, char *dst, size_t size)
{
	blake2b_state ctx;
	uint32_t i;
	uint32_t IV;	/* host byte order independend hash IV */

	int fullblocks = size / BLAKE2B_OUTBYTES;
	int padding = size % BLAKE2B_OUTBYTES;

	/* hash block the whole data set with different IVs to produce
	 * more than just a single data block
	 */
	for (i = 0; i < fullblocks; i++) {
		if (blake2b_init(&ctx, BLAKE2B_OUTBYTES)) return -1;
		IV = htonl(i);
		if (blake2b_update(&ctx, (const char *)&IV, sizeof(IV))) return -1;
		if (blake2b_update(&ctx, src, size)) return -1;
		if (blake2b_final(&ctx, dst+BLAKE2B_OUTBYTES*i, BLAKE2B_OUTBYTES)) return -1;
	}

	if (padding) {
		if (blake2b_init(&ctx, padding)) return -1;
		IV = htonl(i);
		if (blake2b_update(&ctx, (const char *)&IV, sizeof(IV))) return -1;
		if (blake2b_update(&ctx, src, size)) return -1;
		if (blake2b_final(&ctx, dst+BLAKE2B_OUTBYTES*i, padding)) return -1;
	}
	return 0;
}

/*
 * Information splitting. The amount of data is multiplied by
 * blocknumbers. The same blocksize and blocknumbers values
 * must be supplied to AF_merge to recover information.
 * return 0 on success.
 */
int AF_split(char *src, char *dst, size_t blocksize, size_t blocknumbers)
{
	size_t i;
	char *bufblock;
	int r = 0;

	if (blocknumbers < 2) return -EINVAL;
	if ((bufblock = calloc(blocksize, 1)) == NULL) return -ENOMEM;

	/* process everything except the last block */
	for (i = 0; i < blocknumbers-1; i++) {
		charcrandom_buf(dst+(blocksize*i), blocksize);
		XORblock(dst+(blocksize*i), bufblock, bufblock, blocksize);
		if (diffuse(bufblock, bufblock, blocksize)) {
			r = -EINVAL;
			goto out;
		}
	}
	/* the last block is computed */
	XORblock(src, bufblock, dst+(i*blocksize), blocksize);
out:
	secure_memset(bufblock, blocksize);
	free(bufblock);
	return r;
}

/* return 0 on success. */
int AF_merge(char *src, char *dst, size_t blocksize, size_t blocknumbers)
{
	size_t i;
	char *bufblock;
	int r = 0;

	if (blocknumbers < 2) return -EINVAL;
	if ((bufblock = calloc(blocksize, 1)) == NULL) return -ENOMEM;

	for (i = 0; i < blocknumbers-1; i++) {
		XORblock(src+(blocksize*i), bufblock, bufblock, blocksize);
		if (diffuse(bufblock, bufblock, blocksize)) {
			r = -EINVAL;
			goto out;
		}
	}
	XORblock(src+(i*blocksize), bufblock, dst, blocksize);
out:
	secure_memset(bufblock, blocksize);
	free(bufblock);
	return r;
}

